/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.51 from the
 * contents of probe.xs. Do not edit this file, edit probe.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "probe.xs"
#define PERL_NO_GET_CONTEXT     /* we want efficiency */
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "ppport.h"

#define PROBE_TYPE_NONE      0
#define PROBE_TYPE_ONCE      1
#define PROBE_TYPE_PERMANENT 2

#define PROBE_ACTION_LOOKUP  0
#define PROBE_ACTION_REMOVE  1

#define PROBE_SETTINGS_TYPE_INDEX 0
#define PROBE_SETTINGS_ARGS_INDEX 1

/*
 * Use preprocessor macros for time-sensitive operations.
 */
#define probe_is_enabled() !!probe_enabled

static Perl_ppaddr_t probe_nextstate_orig = 0;
static int probe_installed = 0;
static int probe_enabled = 0;
static HV* probe_hash = 0;
static SV* probe_trigger_cb = 0;

static void probe_enable(void);
static void probe_disable(void);
static int probe_is_installed(void);
static void probe_install(pTHX);
static void probe_remove(pTHX);

#define DEBUG 0

#define INFO(x) do { if (DEBUG > 0) dbg_printf x; } while (0)
#define TRACE(x) do { if (DEBUG > 1) dbg_printf x; } while (0)

void dbg_printf(const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
}

static void probe_invoke_callback(pTHX_ const char* file, int line, SV* user_arg, SV* callback)
{
    int count;

    dSP;

    ENTER;
    SAVETMPS;

    PUSHMARK (SP);
    EXTEND(SP, user_arg ? 3 : 2);
    XPUSHs(sv_2mortal(newSVpv(file, 0)));
    XPUSHs(sv_2mortal(newSViv(line)));
    if (user_arg) {
        XPUSHs(user_arg);
    }
    PUTBACK;

    count = call_sv(callback, G_VOID|G_DISCARD);
    if (count != 0) {
        croak("probe trigger should have zero return values");
    }

    FREETMPS;
    LEAVE;
}

static bool probe_lookup(pTHX_ const char* file, int line, int action)
{
    U32 klen = strlen(file);
    char kstr[20];
    SV** rlines = 0;
    HV* lines = 0;

    rlines = hv_fetch(probe_hash, file, klen, 0);
    if (rlines) {
        lines = (HV*) SvRV(*rlines);
        TRACE(("PROBE found entry for file [%s]: %p\n", file, lines));
    } else {
        return false;
    }

    klen = sprintf(kstr, "%d", line);
    if (hv_exists(lines, kstr, klen)) {
        if (action == PROBE_ACTION_REMOVE) {
            /* TODO: remove file name when last line for file was removed? */
            hv_delete(lines, kstr, klen, G_DISCARD);
            INFO(("PROBE removed entry for line [%s]\n", kstr));
        }
        return true;
    } else {
        return false;
    }

    /* catch any mistakes */
    return false;
}

static AV* probe_settings(pTHX_ const char* file, int line)
{
    U32 klen = strlen(file);
    char kstr[20];
    SV** rlines = 0;
    HV* lines = 0;
    SV** rsettings = 0;
    AV* settings = 0;

    rlines = hv_fetch(probe_hash, file, klen, 0);
    if (!rlines) {
        return 0;
    }
    lines = (HV*) SvRV(*rlines);

    klen = sprintf(kstr, "%d", line);
    rsettings = hv_fetch(lines, kstr, klen, 0);
    if (!rsettings) {
        return 0;
    }

    if (!SvROK(*rsettings) || SvTYPE(SvRV(*rsettings)) != SVt_PVAV) {
        croak("Devel::Probe settings must be an ARRAY ref");
    }

    settings = (AV*) SvRV(*rsettings);
    return settings;
}

/*
 * This function will run for every single line in your Perl code.
 * You would do well to make it as cheap as possible.
 */
static OP* probe_nextstate(pTHX)
{
    OP* ret = probe_nextstate_orig(aTHX);

    do {
        const char* file = 0;
        int line = 0;
        AV* settings = 0;
        SV* user_callback_arg = 0;
        int type = PROBE_TYPE_NONE;

        if (!probe_is_enabled()) {
            break;
        }

        file = CopFILE(PL_curcop);
        line = CopLINE(PL_curcop);
        TRACE(("PROBE check [%s] [%d]\n", file, line));
        if (!probe_lookup(aTHX_ file, line, PROBE_ACTION_LOOKUP)) {
            break;
        }

        settings = probe_settings(aTHX_ file, line);
        if (!settings) {
            break;
        }

        type = SvIV(*(av_fetch(settings, PROBE_SETTINGS_TYPE_INDEX, 0)));
        if (av_top_index(settings) == PROBE_SETTINGS_ARGS_INDEX) {
            user_callback_arg = *(av_fetch(settings, PROBE_SETTINGS_ARGS_INDEX, 0));
        }

        INFO(("PROBE triggered [%s] [%d] [%d]\n", file, line, type));
        if (probe_trigger_cb) {
            probe_invoke_callback(aTHX_ file, line, user_callback_arg, probe_trigger_cb);
        }

        if (type == PROBE_TYPE_ONCE) {
            probe_lookup(aTHX_ file, line, PROBE_ACTION_REMOVE);
        }
    } while (0);

    return ret;
}

static void probe_enable(void)
{
    if (probe_is_enabled()) {
        return;
    }
    INFO(("PROBE enabling\n"));
    probe_enabled = 1;
}

static void probe_clear(pTHX)
{
    if (probe_hash) {
        hv_clear(probe_hash);
    } else {
        probe_hash = newHV();
    }
    INFO(("PROBE cleared\n"));
}

static void probe_reset(pTHX_ int installed)
{
    probe_installed = installed;
    probe_enabled = 0;
    probe_clear(aTHX);
    if (probe_trigger_cb) {
        SvREFCNT_dec(probe_trigger_cb);
    }
    probe_trigger_cb = 0;
}

static void probe_disable(void)
{
    if (!probe_is_enabled()) {
        return;
    }
    probe_enabled = 0;
    INFO(("PROBE disabled\n"));
}

static int probe_is_installed(void)
{
    return probe_installed;
}

static void probe_install(pTHX)
{
    if (probe_is_installed()) {
        return;
    }

    INFO(("PROBE installed, [%p] => [%p]\n", PL_ppaddr[OP_NEXTSTATE], probe_nextstate));

    if (!probe_nextstate_orig) {
        probe_nextstate_orig = PL_ppaddr[OP_NEXTSTATE];
    }
    PL_ppaddr[OP_NEXTSTATE] = probe_nextstate;
    probe_reset(aTHX_ 1);
    probe_clear(aTHX);
}

static void probe_remove(pTHX)
{
    if (!probe_is_installed()) {
        return;
    }
    INFO(("PROBE removed, [%p] => [%p]\n", PL_ppaddr[OP_NEXTSTATE], probe_nextstate_orig));
    if (probe_nextstate_orig) {
        PL_ppaddr[OP_NEXTSTATE] = probe_nextstate_orig;
    }
    probe_reset(aTHX_ 0);
}

#line 265 "probe.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 409 "probe.c"

XS_EUPXS(XS_Devel__Probe_install); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_install)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 263 "probe.xs"
    probe_install(aTHX);
#line 420 "probe.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Devel__Probe_remove); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_remove)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 268 "probe.xs"
    probe_remove(aTHX);
#line 435 "probe.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Devel__Probe_is_installed); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_is_installed)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
	int	RETVAL;
	dXSTARG;
#line 273 "probe.xs"
    RETVAL = probe_is_installed();
#line 452 "probe.c"
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Devel__Probe_enable); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_enable)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 279 "probe.xs"
    probe_enable();
#line 469 "probe.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Devel__Probe_disable); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_disable)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 284 "probe.xs"
    probe_disable();
#line 484 "probe.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Devel__Probe_is_enabled); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_is_enabled)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
	int	RETVAL;
	dXSTARG;
#line 289 "probe.xs"
    RETVAL = probe_is_enabled();
#line 501 "probe.c"
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Devel__Probe_clear); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_clear)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 295 "probe.xs"
    probe_disable();
    probe_clear(aTHX);
#line 519 "probe.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Devel__Probe__internal_probe_state); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe__internal_probe_state)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
	HV *	RETVAL;
#line 301 "probe.xs"
    RETVAL = probe_hash;
#line 535 "probe.c"
	{
	    SV * RETVALSV;
	    RETVALSV = newRV((SV*)RETVAL);
	    RETVALSV = sv_2mortal(RETVALSV);
	    ST(0) = RETVALSV;
	}
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Devel__Probe_trigger); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Devel__Probe_trigger)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "callback");
    {
	SV*	callback = ST(0)
;
#line 307 "probe.xs"
    if (probe_trigger_cb == (SV*)NULL) {
        probe_trigger_cb = newSVsv(callback);
    } else {
        SvSetSV(probe_trigger_cb, callback);
    }
#line 562 "probe.c"
    }
    XSRETURN_EMPTY;
}

#ifdef __cplusplus
extern "C" {
#endif
XS_EXTERNAL(boot_Devel__Probe); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Devel__Probe)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("Devel::Probe::install", XS_Devel__Probe_install);
        newXS_deffile("Devel::Probe::remove", XS_Devel__Probe_remove);
        newXS_deffile("Devel::Probe::is_installed", XS_Devel__Probe_is_installed);
        newXS_deffile("Devel::Probe::enable", XS_Devel__Probe_enable);
        newXS_deffile("Devel::Probe::disable", XS_Devel__Probe_disable);
        newXS_deffile("Devel::Probe::is_enabled", XS_Devel__Probe_is_enabled);
        newXS_deffile("Devel::Probe::clear", XS_Devel__Probe_clear);
        newXS_deffile("Devel::Probe::_internal_probe_state", XS_Devel__Probe__internal_probe_state);
        newXS_deffile("Devel::Probe::trigger", XS_Devel__Probe_trigger);
#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

#ifdef __cplusplus
}
#endif
